package algo.util

import algo.util.Matrix.{MATRICES_PATH, MATRIX_FILE_ENDING}

import java.io.{File, PrintWriter}
import scala.io.Source

case class Matrix(data: Seq[Seq[Int]], name: String) {
  def mul(thatInverted: Matrix, cache: Boolean = true): Matrix = {
    val fileName = this.name + "_mul_" + thatInverted.name
    if (cache) {
      Matrix.fromCSV(fileName) match {
        case Some(matrix) => matrix
        case None =>
          val newMatrix = Matrix(mul(this.data, thatInverted.data), fileName)
          newMatrix.toCSV()
          newMatrix
      }
    } else {
      Matrix(mul(this.data, thatInverted.data), fileName)
    }
  }

  private def mul(A: Seq[Seq[Int]], Binv: Seq[Seq[Int]]): Seq[Seq[Int]] = mul(A.map(_.toArray[Int]).toArray[Array[Int]], Binv.map(_.toArray[Int]).toArray[Array[Int]])
  private def mul(A: Array[Array[Int]], Binv: Array[Array[Int]]): Seq[Seq[Int]] = {
    assert(A(0).length == Binv(0).length)
    val result: Array[Array[Int]] = Array.ofDim[Int](A.length, Binv.length)

    // use for loop to allow scala to automatically parallelize the computation (with 'map' it would take ages)
    for (i <- A.indices; j <- Binv.indices) {
      var accum = 0
      for (k <- A(0).indices) {
        accum += A(i)(k) * Binv(j)(k)
      }
      result(i)(j) = accum
    }
    result.map(_.toSeq).toSeq
  }

  def tiledMul(thatInverted: Matrix, tileSize: Int, cache: Boolean = true): Matrix = {
    val fileName = this.name + "_mul_" + thatInverted.name
    if (cache) {
      Matrix.fromCSV(fileName) match {
        case Some(matrix) => matrix
        case None =>
          val newMatrix = Matrix(tiledMul(this.data, thatInverted.data, tileSize), fileName)
          newMatrix.toCSV()
          newMatrix
      }
    } else {
      Matrix(tiledMul(this.data, thatInverted.data, tileSize), fileName)
    }
  }

  private def tiledMul(A: Seq[Seq[Int]], Binv: Seq[Seq[Int]], tileSize: Int): Seq[Seq[Int]] = tiledMul(A.map(_.toArray[Int]).toArray[Array[Int]], Binv.map(_.toArray[Int]).toArray[Array[Int]], tileSize)
  private def tiledMul(A: Array[Array[Int]], Binv: Array[Array[Int]], tileSize: Int): Seq[Seq[Int]] = {
    assert(A(0).length == Binv(0).length)
    val result: Array[Array[Int]] = Array.ofDim[Int](A.length, Binv.length)

    val numOfATiles = A.length / tileSize
    val numOfBinvTiles = Binv.length / tileSize
    assert(tileSize % numOfBinvTiles == 0)
    // use for loop to allow scala to automatically parallelize the computation (with 'map' it would take ages)

    for (i <- 0 until numOfATiles; j <- 0 until numOfBinvTiles) {
      for (ii <- 0 until tileSize; jj <- 0 until tileSize) {
        var accum = 0
        for (k <- A(0).indices) {
          accum += A(i * tileSize + ii)(k) * Binv(j * tileSize + jj)(k)
        }
        result(i * tileSize + (ii / numOfBinvTiles + j * tileSize / numOfBinvTiles))((ii % numOfBinvTiles) * tileSize + jj) = accum
      }
    }
    result.map(_.toSeq).toSeq
  }

  def add(that: Matrix, cache: Boolean = true): Matrix = {
    val fileName = this.name + "_add_" + that.name
    if (cache) {
      Matrix.fromCSV(fileName) match {
        case Some(matrix) => matrix
        case None =>
          val newMatrix = Matrix(add(this.data, that.data), fileName)
          newMatrix.toCSV()
          newMatrix
      }
    } else {
      Matrix(add(this.data, that.data), fileName)
    }
  }

  private def add(A: Seq[Seq[Int]], B: Seq[Seq[Int]]): Seq[Seq[Int]] = add(A.map(_.toArray[Int]).toArray[Array[Int]], B.map(_.toArray[Int]).toArray[Array[Int]])
  private def add(A: Array[Array[Int]], B: Array[Array[Int]]): Seq[Seq[Int]] = {
    assert(A(0).length == B(0).length)
    assert(A(1).length == B(1).length)
    val result: Array[Array[Int]] = Array.ofDim[Int](A.length, B.length)

    // use for loop to allow scala to automatically parallelize the computation (with 'map' it would take ages)
    for (i <- A.indices; j <- B.indices) {
      result(i)(j) = A(i)(j) + B(i)(j)
    }
    result.map(_.toSeq).toSeq
  }

  def pad(topSize: Int, bottomSize: Int, leftSize: Int, rightSize: Int, cache: Boolean = true): Matrix = {
    val fileName = this.name + "_pad_" + topSize.toString + "_" + bottomSize.toString + "_" + leftSize.toString + "_" + rightSize.toString
    if (cache) {
      Matrix.fromCSV(fileName) match {
        case Some(matrix) => matrix
        case None =>
          val newMatrix = Matrix(pad(this.data, topSize, bottomSize, leftSize, rightSize), fileName)
          newMatrix.toCSV()
          newMatrix
      }
    } else {
      Matrix(pad(this.data, topSize, bottomSize, leftSize, rightSize), fileName)
    }
  }

  private def pad(A: Seq[Seq[Int]], topSize: Int, bottomSize: Int, leftSize: Int, rightSize: Int): Seq[Seq[Int]] = pad(A.map(_.toArray[Int]).toArray[Array[Int]], topSize, bottomSize, leftSize, rightSize)
  private def pad(A: Array[Array[Int]], topSize: Int, bottomSize: Int, leftSize: Int, rightSize: Int): Seq[Seq[Int]] = {

    val result: Array[Array[Int]] = Array.ofDim[Int](topSize + A.length + bottomSize, leftSize + A(0).length + rightSize)
    for (i <- result.indices) {
      for (j <- result(0).indices) {
        if(i < topSize || i > result.length - 1 - bottomSize || j < leftSize || j > result(0).length - 1 - rightSize) {
          result(i)(j) = 0
        } else {
          result(i)(j) = A(i - topSize)(j - leftSize)
        }
      }
    }
    result.map(_.toSeq).toSeq
  }

  def conv(transposedWeight: Matrix, inChannel: Int, cache: Boolean = true): Matrix = {
    val fileName = this.name + "_mul_" + transposedWeight.name + "_" + inChannel.toString
    if (cache) {
      Matrix.fromCSV(fileName) match {
        case Some(matrix) => matrix
        case None =>
          val newMatrix = Matrix(conv(this.data, transposedWeight.data, inChannel), fileName)
          newMatrix.toCSV()
          newMatrix
      }
    } else {
      Matrix(conv(this.data, transposedWeight.data, inChannel), fileName)
    }
  }

  private def conv(A: Seq[Seq[Int]], transposedWeight: Seq[Seq[Int]], inChannel: Int): Seq[Seq[Int]] = conv(A.map(_.toArray[Int]).toArray[Array[Int]], transposedWeight.map(_.toArray[Int]).toArray[Array[Int]], inChannel)
  private def conv(A: Array[Array[Int]], transposedWeight: Array[Array[Int]], inChannel: Int): Seq[Seq[Int]] = {
    val wSize = math.sqrt(transposedWeight(0).length / inChannel)
    assert(wSize % 1 == 0 && A(0).length % inChannel == 0 && A.length > wSize && A(0).length / inChannel > wSize)
    val newIWidth = (A(0).length / inChannel).toInt - wSize.toInt + 1
    val result: Array[Array[Int]] = Array.ofDim[Int](A.length - wSize.toInt + 1, newIWidth * transposedWeight.length)

    // use for loop to allow scala to automatically parallelize the computation (with 'map' it would take ages)
    for (i <- result.indices) {
      for (j <- 0 to newIWidth - 1) {
        for (k <- transposedWeight.indices) {
          var accum = 0
          for(m <- 0 to wSize.toInt - 1) {
            for(n <- 0 to wSize.toInt - 1) {
              for(l <- 0 to inChannel - 1) {
                accum += A(i + n)((j + m) * inChannel + l) * transposedWeight(k)(m * wSize.toInt * inChannel + n * inChannel + l)
              }
            }
          }
          result(i)(j * transposedWeight.length + k) = accum
        }
      }
    }
    result.map(_.toSeq).toSeq
  }

  def sumRows(): Matrix = Matrix(Seq(data.map(_.sum)), this.name + "_sumrows")

  def inc(): Matrix = Matrix(data.map(_.map(_ + 1)), name + "_inc")

  def mod(divisor: Int): Matrix = Matrix(data.map(_.map(_ % divisor)), name + "_mod_" + divisor.toInt)

  def toCSV(): Unit = {
    val folder = new File(MATRICES_PATH)
    if (!folder.exists() || !folder.isDirectory) {
      folder.mkdirs
    }
    val file = new PrintWriter(new File(MATRICES_PATH + name + MATRIX_FILE_ENDING))
    file.write(data.map(_.mkString(", ")).mkString("\n"))
    file.close()
  }
}

object Matrix {

  final val MATRICES_PATH = "resources/matrices/"
  final val MATRIX_FILE_ENDING = ".csv"

  def fromCSV(fileName: String): Option[Matrix] = {
    val fullPath = MATRICES_PATH + fileName + MATRIX_FILE_ENDING
    if (!new File(fullPath).exists()) {
      None
    } else {
      val file = Source.fromFile(fullPath)
      val matrix = Matrix(file.getLines().map(_.split(",").map(_.filter(_.isDigit).toInt).toList).toList, fileName)
      file.close()
      Some(matrix)
    }
  }

  def fromRandom(rows: Int, cols: Int, maxInt: Int = 256, cacheId: Option[String] = None): Matrix = {
    val r = scala.util.Random
    val fileName = cacheId match {
      case Some(c) => s"matrix_${rows}_${cols}_random${maxInt}_${c}"
      case None => s"matrix_${rows}_${cols}_random${maxInt}"
    }
    if (cacheId.isDefined) {
      fromCSV(fileName) match {
        case Some(matrix) => matrix
        case None =>
          val newMatrix = Matrix(Seq.fill(rows)(Seq.fill(cols)(r.nextInt(maxInt))), fileName)
          newMatrix.toCSV()
          newMatrix
      }
    } else {
      Matrix(Seq.fill(rows)(Seq.fill(cols)(r.nextInt(maxInt))), fileName)
    }
  }

  def fromCounter(rows: Int, cols: Int, repetitions: Int = 1, cacheId: Option[String] = None): Matrix = {
    val fileName = cacheId match {
      case Some(c) => s"matrix_${rows}_${cols}_counter_rep${repetitions}_${c}"
      case None => s"matrix_${rows}_${cols}_counter_rep${repetitions}"
    }
    if (cacheId.isDefined) {
      fromCSV(fileName) match {
        case Some(matrix) => matrix
        case None =>
          val newMatrix = Matrix(Seq.fill(rows)((1 to (cols / repetitions)).flatMap(Seq.fill(repetitions)(_))), fileName)
          newMatrix.toCSV()
          newMatrix
      }
    } else {
      Matrix(Seq.fill(rows)((1 to (cols / repetitions)).flatMap(Seq.fill(repetitions)(_))), fileName)
    }
  }

}
