package backend.hdl.arch

import backend.hdl.arch.device.DeviceSpecificCompiler
import backend.hdl.arch.mem.WriteAsync
import backend.hdl.arch.rewrite.{MergeIntoCounterRules, MoveDownRepeatHiddenRules, MoveDownRepeatRules, RemoveSelectRules}
import backend.hdl._
import core._
import core.compile.{CompilerPass, CompilerPhase}
import core.rewrite.{RewriteAll, RewriteStep}
import core.util.IRDotGraph

object MapCompiler extends CompilerPass {

  override def phaseBefore: CompilerPhase = DeviceSpecificCompiler.phaseAfter

  override def run(expr: Expr): Expr = {
    val newExpr0 = TypeChecker.checkAssert(expr)
    val newExpr1 = TypeChecker.check(removeUnboundParams(newExpr0))
    val newExpr2 = RewriteStep(RewriteAll(), MoveDownRepeatRules.get() ++ RemoveSelectRules.get()).apply(newExpr1)
    val newExpr3 = TypeChecker.check(repeatFunctionLeaves(newExpr2))
    val newExpr4 = RewriteStep(RewriteAll(), MoveDownRepeatHiddenRules.get()).apply(newExpr3)
    // assert that the RepeatHidden have been properly propagated
    newExpr4.visit({
      case InverseRepeatHidden(_, _, _) => throw MalformedExprException("temporary helper node found after rewriting! This should not happen - maybe a rewrite rule is missing?")
      case _ =>
    })
    val newExpr5 = RewriteStep(RewriteAll(), MergeIntoCounterRules.get()).apply(newExpr4)
    val newExpr6 = TypeChecker.check(replaceMapStreams(newExpr5))
//    IRDotGraph(newExpr6).show()
//    IRDotGraphAnimation(Seq(newExpr0, newExpr5)).show()
    newExpr6
  }

  private def removeUnboundParams(expr: Expr): Expr = expr match {
    case MapOrderedStream(function: LambdaT, input, t: OrderedStreamTypeT) =>
      val unboundParams = getUnboundParams(function).distinct
      assert(unboundParams.size <= 2) // TODO for now, code only works with 1 free param

      if (unboundParams.isEmpty) {
        MapOrderedStream(
          removeUnboundParams(function),
          removeUnboundParams(input)
        )
      } else if(unboundParams.size == 2) {
        MapOrderedStream({
          val newParam = ParamDef()
          ArchLambda(
            newParam,
            // Important: first treat children (here: body) and then the current node. Traverse from leaves to the top.
            // Otherwise a more complex nesting of Select primitives would be generated!
            removeUnboundParams(function.body).visitAndRebuild{
              case ParamUse(pd) if pd.id == function.param.id =>
                Select3(ParamUse(newParam), 0)
              case pu: ParamUse if unboundParams.contains(pu) =>
                Select3(ParamUse(newParam), unboundParams.indexOf(pu) + 1)
              case e => e
            }.asInstanceOf[Expr]
          )},
          removeUnboundParams(Zip3OrderedStream(Tuple3(input, Repeat(unboundParams.head, t.len), Repeat(unboundParams(1), t.len))))
        )
      } else {
        MapOrderedStream({
          val newParam = ParamDef()
          ArchLambda(
            newParam,
            // Important: first treat children (here: body) and then the current node. Traverse from leaves to the top.
            // Otherwise a more complex nesting of Select primitives would be generated!
            removeUnboundParams(function.body).visitAndRebuild{
              case ParamUse(pd) if pd.id == function.param.id =>
                Select(ParamUse(newParam), 0)
              case pu: ParamUse if unboundParams.contains(pu) =>
                Select(ParamUse(newParam), unboundParams.indexOf(pu) + 1)
              case e => e
            }.asInstanceOf[Expr]
          )},
          // TODO change for multiple unbound vars
          removeUnboundParams(ZipOrderedStream(Tuple(input, Repeat(unboundParams.head, t.len))))
        )
      }
    case MapUnorderedStream(function, input, t: UnorderedStreamTypeT) =>
      ???
    // Reduce -based primitives do not need this mechanism because there is no need to ever repeat their input stream
    case _ => expr.build(expr.children.map({
      case e: Expr => removeUnboundParams(e)
      case t: Type => t
    }))
  }

  private def repeatFunctionLeaves(expr: Expr): Expr = expr.visitAndRebuild{
    case MapOrderedStream(ArchLambda(p, body, _), input, _) =>
      val OrderedStreamType(_, inputLength) = TypeChecker.checkIfUnknown(input, input.t).t
      MapOrderedStream(
        ArchLambda(p,
          RepeatHidden(
            body.visitAndRebuild{
              case ParamUse(pd) if pd.id == p.id => InverseRepeatHidden(ParamUse(pd), inputLength)
              case e => e
          }.asInstanceOf[Expr], inputLength)
        ),
        input
      )
    case MapUnorderedStream(ArchLambda(p, body, _), input, _) =>
      val UnorderedStreamType(_, inputLength) = TypeChecker.checkIfUnknown(input, input.t).t
      MapUnorderedStream(
        ArchLambda(p,
          RepeatHidden(
            body.visitAndRebuild{
              case ParamUse(pd) if pd.id == p.id => InverseRepeatHidden(ParamUse(pd), inputLength)
              case e => e
            }.asInstanceOf[Expr], inputLength)
        ),
        input
      )
    case FoldOrderedStream(ArchLambda(p1, ArchLambda(p2, body, _), _), input, _) =>
      val OrderedStreamType(_, inputLength) = TypeChecker.checkIfUnknown(input, input.t).t
      FoldOrderedStream(
        ArchLambda(p1, ArchLambda(p2,
          RepeatHidden(
            body.visitAndRebuild{
              case ParamUse(pd) if pd.id == p1.id || pd.id == p2.id => InverseRepeatHidden(ParamUse(pd), inputLength)
              case e => e
            }.asInstanceOf[Expr], inputLength)
        )),
        input
      )
    case ReduceOrderedStream(ArchLambda(p1, ArchLambda(p2, body, _), _), initialValue, input, _) =>
      val OrderedStreamType(_, inputLength) = TypeChecker.checkIfUnknown(input, input.t).t
      ReduceOrderedStream(
        ArchLambda(p1, ArchLambda(p2,
          RepeatHidden(
            body.visitAndRebuild{
              case ParamUse(pd) if pd.id == p1.id || pd.id == p2.id => InverseRepeatHidden(ParamUse(pd), inputLength)
              case e => e
            }.asInstanceOf[Expr], inputLength)
        )),
        initialValue,
        input
      )
    case e =>e
  }.asInstanceOf[Expr]

  private def replaceMapStreams(expr: Expr): Expr = expr match {
    // maps within repeat or loop expressions may be repeated and cannot be replaced by simplified implementations
    case repeat @ Repeat(_, _, _) => skipMapStreams(repeat)
    case repeatHidden @ RepeatHidden(_, _, _) => skipMapStreams(repeatHidden)
    case MapOrderedStream(function: LambdaT, input, _) =>
      val unboundParams = getUnboundParams(function)
      (unboundParams.isEmpty, hasRepeatedStreamInput(function)) match {
        case (true, false) =>
          MapSimpleOrderedStream(
            replaceMapStreams(function),
            replaceMapStreams(input)
          )
        case (_, true) =>
          MapOrderedStream(
            replaceMapStreams(function),
            input
          )
        case _ =>
          MapOrderedStream(
            replaceMapStreams(function),
            replaceMapStreams(input)
          )
      }
    case MapUnorderedStream(function, input, t: UnorderedStreamTypeT) =>
      ???
      // TODO fix vhdl templates first!
    case _ => expr.build(expr.children.map({
      case e: Expr => replaceMapStreams(e)
      case t: Type => t
    }))
  }

  private def skipMapStreams(expr: Expr): Expr = expr match {
    /* reduction-like expressions (like fold, reduce, writeAsync (remember WriteAsync takes a stream of data and returns a base address))
    return a single element (and not a stream anymore); and single elements can be repeated without any problem.
    No need to 'skip' MapStreams any longer, go back to 'replaceMapStreams'. */
    case fold @ FoldOrderedStream(_, _, _) => replaceMapStreams(fold)
    case reduce @ ReduceOrderedStream(_, _, _, _) => replaceMapStreams(reduce)
    case writeAsync @ WriteAsync(_, _, _, _) => replaceMapStreams(writeAsync)
    case _ => expr.build(expr.children.map({
      case e: Expr => skipMapStreams(e)
      case t: Type => t
    }))
  }

  private def getUnboundParams(expr: Expr, paramIds: Seq[Long] = Seq()): Seq[ParamUse] = expr match {
    case l: LambdaT => getUnboundParams(l.body, paramIds :+ l.param.id)
    // do not collect shared functions! We do not want to 'repeat' anything in shared functions!
    case pu @ ParamUse(pd) if !pu.t.isInstanceOf[FunTypeT] && !paramIds.contains(pd.id) => Seq(pu)
    case _ => expr.children.flatMap{
      case e: Expr => getUnboundParams(e, paramIds)
      case _ => Seq()
    }
  }

  private def hasRepeatedStreamInput(lambda: LambdaT): Boolean = {
    lambda.visit{
      case Repeat(input, _, _) if {
        (input match {
          case Repeat(_, _, _) => false
          case RepeatHidden(_, _, _) => false
          case _ => true
        }) &&
        (input.t match {
          case _: UnorderedStreamTypeT => true
          case _ => false // 1D data can easily be repeated without any repetition related problems
        }) &&
        hasParamUse(input, lambda.param.id)
      } => return true
      case _ =>
    }
    false
  }

  private def hasParamUse(expr: Expr, paramId: Long): Boolean = {
    expr.visit{
      case ParamUse(pd) if pd.id == paramId => return true
      case _ =>
    }
    false
  }
}

case class InverseRepeatHiddenExpr private[arch](innerIR: Expr, types: Seq[Type]) extends BuiltinExpr {
  override def build(newInnerIR: Expr, newTypes: Seq[Type]): InverseRepeatHiddenExpr = InverseRepeatHiddenExpr(newInnerIR, newTypes)
}
object InverseRepeatHidden extends BuiltinExprFun {
  override def args: Int = 1
  private[arch] def apply(input: Expr, dimensions: Seq[ArithTypeT]): Expr = dimensions.size match {
    case 0 => input
    case _ => InverseRepeatHidden(apply(input, dimensions.tail), dimensions.head)
  }
  def apply(input: Expr, len: ArithTypeT): InverseRepeatHiddenExpr = InverseRepeatHiddenExpr({
    val stv = NonArrayTypeVar()
    FunctionCall(TypeFunctionCall(BuiltinTypeFunction(TypeFunType(stv, HWFunType(stv, stv))), input.t), input)
  }, Seq(len))
  override protected def apply(inputs: Seq[Expr], types: Seq[Type]): InverseRepeatHiddenExpr = apply(inputs.head, types.head.asInstanceOf[ArithTypeT])
  def unapply(expr: InverseRepeatHiddenExpr): Some[(Expr, ArithTypeT, Type)] =
    Some(expr.args.head, expr.types.head.asInstanceOf[ArithTypeT], expr.t)
}
