package backend.hdl.arch.rewrite

import backend.hdl.NamedTupleTypeT
import backend.hdl.arch.mem.{ReadAsyncOrdered, ReadSync}
import backend.hdl.arch._
import core.rewrite.{Rule, RulesT}
import core._

object MoveDownRepeatRules extends RulesT {

  override protected def all(config: Option[Int]): Seq[Rule] =
    Seq(moveDownMapRepeatParamUse, moveRepeatIntoLet, moveMapRepeatIntoLet, moveDownRepeat2, moveDownRepeat3, moveDownRepeat4, reorderRepeats, removeRepeatTupleType, moveDownRepeatIntoReadAddress)//, removeRepeat)

  private def moveDownMapRepeatParamUse: Rule = Rule("moveDownMapRepeatParamUse", {
    case MapOrderedStream(ArchLambda(param, body, _), input, _) if {
      // if all the ParamUses of the function's parameter are directly wrapped in a Repeat with the same N
      var reps: Seq[ArithTypeT] = Seq()
      var paramUses: Int = 0
      body.visit{
        case Repeat(ParamUse(pd), repetitions, _) if pd.id == param.id =>
          reps = reps :+ repetitions
        case ParamUse(pd) if pd.id == param.id =>
          paramUses = paramUses + 1
        case _ =>
      }
      // prevent loops; check if this rewrite rule has already been applied
      val isMapRepeatOnly: Boolean = body match {
        case Repeat(ParamUse(_), _, _) => true
        case _ => false
      }
      !isMapRepeatOnly && reps.nonEmpty && reps.forall(_ == reps.head) && paramUses == reps.size
    } =>
      var reps: ArithTypeT = 0
      val newFun = {
        val newParam = ParamDef()
        ArchLambda(
          newParam,
          body.visitAndRebuild{
            case Repeat(ParamUse(pd), repetitions, _) if pd.id == param.id => // Avoid removing multiple repeats
              reps = repetitions
              ParamUse(newParam)
            case e => e
          }.asInstanceOf[Expr]
        )
      }

      MapOrderedStream(
        newFun,
        MapOrderedStream(Repeat.asFunction(Seq(None), Seq(reps)), input)
      )
  })

  private def moveRepeatIntoLet: Rule = Rule("moveRepeatIntoLet", {
    case Let(param, body, arg, _) if {
      // if all the ParamUses of the function's parameter are directly wrapped in a Repeat with the same N
      var reps: Seq[ArithTypeT] = Seq()
      var paramUses: Int = 0
      body.visit{
        case Repeat(ParamUse(pd), repetitions, _) if pd.id == param.id =>
          reps = reps :+ repetitions
        case ParamUse(pd) if pd.id == param.id =>
          paramUses = paramUses + 1
        case _ =>
      }
      reps.nonEmpty && reps.forall(_ == reps.head) && paramUses == reps.size
    } =>
      var reps: ArithTypeT = 0
      val newParam = ParamDef()
      val newBody = {
        body.visitAndRebuild{
          case Repeat(ParamUse(pd), repetitions, _) if pd.id == param.id || pd.id == newParam.id =>
            reps = repetitions
            ParamUse(newParam)
          case e => e
        }.asInstanceOf[Expr]
      }
      val newArg = Repeat(arg, reps)
      Let(newParam, newBody, newArg)
  })

  private def moveMapRepeatIntoLet: Rule = Rule("moveMapRepeatIntoLet", {
    case Let(param, body, arg, _) if {
      // if all the ParamUses of the function's parameter are directly wrapped in a Map-Repeat with the same N
      var reps: Seq[ArithTypeT] = Seq()
      var paramUses: Int = 0
      body.visit{
        case MapOrderedStream(ArchLambda(p1, Repeat(ParamUse(p2), repetitions, _), _), ParamUse(pd), _) if pd.id == param.id && p1.id == p2.id =>
          reps = reps :+ repetitions
        case ParamUse(pd) if pd.id == param.id =>
          paramUses = paramUses + 1
        case _ =>
      }
      reps.nonEmpty && reps.forall(_ == reps.head) && paramUses == reps.size
    } =>
      var reps: ArithTypeT = 0
      val newParam = ParamDef()
      val newBody = {
        body.visitAndRebuild{
          case MapOrderedStream(ArchLambda(p1, Repeat(ParamUse(p2), repetitions, _), _), ParamUse(pd), _) if pd.id == param.id && p1.id == p2.id =>
            reps = repetitions
            ParamUse(newParam)
          case e => e
        }.asInstanceOf[Expr]
      }
      val newArg = MapOrderedStream(Repeat.asFunction(Seq(None), Seq(reps)), arg)
      Let(newParam, newBody, newArg)
  })

  private def moveDownRepeat2: Rule = Rule("moveDownRepeat2", {
    case MapOrderedStream(fun @ ArchLambda(p1, Repeat(ParamUse(p2), _, _), _), RepeatHidden(input, repetitions, _), _) if p1.id == p2.id =>
      RepeatHidden(MapOrderedStream(fun, input), repetitions)
    case MapOrderedStream(fun @ ArchLambda(p1, Repeat(ParamUse(p2), _, _), _), Marker(input, text), _) if p1.id == p2.id =>
      Marker(MapOrderedStream(fun, input), text)
    case Repeat(RepeatHidden(input, hiddenReps, _), reps, _) =>
      RepeatHidden(Repeat(input, reps), hiddenReps)
    case Repeat(MapOrderedStream(fun, input, _), repetitions, _) =>
      MapOrderedStream(MapOrderedStream.asFunction(Seq(Some(fun), None)), Repeat(input, repetitions))
    case Repeat(JoinOrderedStream(input, _), repetitions, _) =>
      MapOrderedStream(JoinOrderedStream.asFunction(), Repeat(input, repetitions))
    case Repeat(Let(param, body, arg, _), repetitions, _) =>
      Let(param, Repeat(body, repetitions), arg)
    case MapOrderedStream(fun @ ArchLambda(p1, Repeat(ParamUse(p2), _, _), _), Let(param, body, arg, _), _) if p1.id == p2.id =>
      Let(param, MapOrderedStream(fun, body), arg)
    case MapOrderedStream(ArchLambda(p1, Repeat(ParamUse(p2), repetitions, _), _), ZipOrderedStream(Tuple(input1, input2, _), _), _) if p1.id == p2.id =>
      MapOrderedStream(
        ZipOrderedStream.asFunction(),
        ZipOrderedStream(Tuple(
          MapOrderedStream(Repeat.asFunction(Seq(None), Seq(repetitions)), input1),
          MapOrderedStream(Repeat.asFunction(Seq(None), Seq(repetitions)), input2)
        ))
      )
    case Repeat(Tuple(input1, input2, _), repetitions, _) =>
      ZipOrderedStream(Tuple(Repeat(input1, repetitions), Repeat(input2, repetitions)))
    case Repeat(Tuple3(input1, input2, input3, _), repetitions, _) =>
      Zip3OrderedStream(Tuple3(Repeat(input1, repetitions), Repeat(input2, repetitions), Repeat(input3, repetitions)))
    case Repeat(ZipOrderedStream(input, _), repetitions, _) =>
      MapOrderedStream(ZipOrderedStream.asFunction(), Repeat(input, repetitions))
    case Repeat(Zip3OrderedStream(input, _), repetitions, _) =>
      MapOrderedStream(Zip3OrderedStream.asFunction(), Repeat(input, repetitions))
    case Repeat(Alternate(f1, f2, input, _), repetitions, _) =>
      MapOrderedStream(Alternate.asFunction(Seq(Some(f1), Some(f2), None)), Repeat(input, repetitions))
    case Repeat(Conversion(input, t), repetitions, _) =>
      MapOrderedStream(
        {
          val param = ParamDef()
          ArchLambda(param, Conversion(ParamUse(param), t))
        },
        Repeat(input, repetitions)
      )
    case MapOrderedStream(ArchLambda(p1, Repeat(ParamUse(p2), times, _), _), Conversion(input, t), _) if p1.id == p2.id =>
      MapOrderedStream(
        {
          val param = ParamDef()
          ArchLambda(param, Conversion(ParamUse(param), t))
        },
        MapOrderedStream(Repeat.asFunction(Seq(None), Seq(times)), input)
      )
  })

  private def moveDownRepeat3: Rule = Rule("moveDownRepeat3", {
    // map fusion
    case MapOrderedStream(ArchLambda(p1, Repeat(ParamUse(p2), repetitions, _), _), MapOrderedStream(ArchLambda(param, body, _), input, _), _) if {
      // Avoid looping with multiple-repeat-removal guard in moveDownMapRepeatParamUse
      val isMapRepeatOnly: Boolean = body match {
        case Repeat(ParamUse(_), _, _) => true
        case _ => false
      }
      !isMapRepeatOnly
    } && p1.id == p2.id =>
      MapOrderedStream(
        {
          ArchLambda(
            param,
            Repeat(body, repetitions)
          )
        },
        input
      )
  })

  private def moveDownRepeat4: Rule = Rule("moveDownRepeat4", {
    case ArchLambda(p1: ParamDef, Repeat(ReadAsyncOrdered(readMemoryController, Select(ParamUse(p2), sel1, _), Select(ParamUse(p3), sel2, _), _), repetitions, _), _) if p1.id == p2.id  && p1.id == p3.id =>
      ArchLambda(
        p1,
        MapOrderedStream(
          {
            val param = ParamDef()
            ArchLambda(
              param,
              ReadAsyncOrdered(
                readMemoryController,
                Select(ParamUse(param), sel1),
                Select(ParamUse(param), sel2)
              )
            )
          },
          Repeat(ParamUse(p1), repetitions)
        )
      )
  })

  private def reorderRepeats: Rule = Rule("reorderRepeats", {
    case MapOrderedStream(ArchLambda(p1: ParamDef, Repeat(ParamUse(p2), reps1, _), _), Repeat(input, reps2, _), _) if p1.id == p2.id =>
      Repeat(Repeat(input, reps1), reps2)
  })

  private def removeRepeatTupleType: Rule = Rule("removeRepeatTupleType", {
    case MapOrderedStream(ArchLambda(p1, body, _), Repeat(input, times, _), _) if input.t.isInstanceOf[NamedTupleTypeT] && {
      var exprFound = false
      body.visit{
        case Select(ParamUse(p2), sel, _) if p1.id == p2.id =>
          exprFound = true
        case _ =>
      }
      exprFound
    } =>
      Repeat(
        body.visitAndRebuild{
          case Select(ParamUse(p2), sel, _) if p1.id == p2.id =>
            Select(input, sel)
          case e => e
        }.asInstanceOf[Expr],
        times
      )
  })

  private def moveDownRepeatIntoReadAddress: Rule = Rule("moveDownRepeatIntoReadAddress", {
    case Repeat(SplitOrderedStream(Let(p1, Let(p2, Let(p3, Let(p4, Let(p5,
    MapOrderedStream(ArchLambda(p6, ReadSync(ParamUse(p7), Select(ParamUse(p8), sel1, _), Select(ParamUse(p9), sel2, _), _), _), ZipOrderedStream(Tuple(rAddr,
    ParamUse(p10), _), _), _) , a5, _), a4, _), a3, _), a2, _), a1, _), chunkSize, _), times, _)  if p5.id == p10.id && p2.id == p7.id && p6.id == p8.id && p6.id == p9.id
    =>
      SplitOrderedStream(SplitOrderedStream(
        Let(p1, Let(p2, Let(p3, Let(p4, Let(p5, MapOrderedStream(ArchLambda(p6,
          ReadSync(ParamUse(p7), Select(ParamUse(p8), sel1), Select(ParamUse(p9), sel2))),
          ZipOrderedStream(Tuple(JoinOrderedStream(Repeat(rAddr, times)), JoinOrderedStream(Repeat(ParamUse(p10), times))))
        ), a5), a4), a3), a2), a1),
        chunkSize), times)

    case MapOrderedStream(ArchLambda(pa, Repeat(ParamUse(pb), times, _), _), SplitOrderedStream(Let(p1, Let(p2, Let(p3, Let(p4, Let(p5,
    MapOrderedStream(ArchLambda(p6, ReadSync(ParamUse(p7), Select(ParamUse(p8), sel1, _), Select(ParamUse(p9), sel2, _), _), _), ZipOrderedStream(Tuple(rAddr,
    ParamUse(p10), _), _), _) , a5, _), a4, _), a3, _), a2, _), a1, _), chunkSize, _), _) if
      p5.id == p10.id && p2.id == p7.id && p6.id == p8.id && p6.id == p9.id && pa.id == pb.id
    =>
      SplitOrderedStream(SplitOrderedStream(
        Let(p1, Let(p2, Let(p3, Let(p4, Let(p5, MapOrderedStream(ArchLambda(p6,
          ReadSync(ParamUse(p7), Select(ParamUse(p8), sel1), Select(ParamUse(p9), sel2))),
          ZipOrderedStream(Tuple(JoinOrderedStream(MapOrderedStream(Repeat.asFunction(Seq(None), Seq(times)), rAddr)),
            JoinOrderedStream(MapOrderedStream(Repeat.asFunction(Seq(None), Seq(times)), ParamUse(p10)))))
        ), a5), a4), a3), a2), a1),
        chunkSize), times)
  })
}
