package backend.hdl.arch.rewrite.conv

import backend.hdl.arch.mem.Read
import backend.hdl.arch.rewrite.RulesHelper
import backend.hdl.arch._
import backend.hdl.{IntType, IntTypeT, OrderedStreamTypeT, VectorTypeT}
import core.rewrite.{Rule, RulesT}
import core._
import lift.arithmetic._

object MoveDownSlideGeneralStreamRules extends RulesT {

  override protected def all(config: Option[Int]): Seq[Rule] =
    Seq(
      skipSplitStream,
      skipVectorToStream,
      skipDropStream,
      skipJoinStream,
      skipMapFun,
      skipMarker,
      moveIntoLet,
      skipMapVecJoinVec,
      skipTransposeVec,
      mapFission,
      mapFissionMapNDSlide,
      mapFusionMapNDSlide,
      mapFusion,
      mergeCounter2D,
    )

  def skipSplitStream: Rule = Rule("moveDownSlideGeneralStream_skipSplitStream", {
    case SlideGeneralOrderedStream(SplitOrderedStream(input, chunkSize, _), windowSize, stepSize, _) =>
      val newWindowSize = windowSize.ae.evalInt * chunkSize.ae.evalInt
      val newStepSize = stepSize.ae.evalInt * chunkSize.ae.evalInt
      val slidedInput = SlideGeneralOrderedStream(input, newWindowSize, newStepSize)
      MapOrderedStream(SplitOrderedStream.asFunction(Seq(None), Seq(chunkSize)), slidedInput)
  })

  def skipVectorToStream: Rule = Rule("moveDownSlideGeneralStream_skipVectorToStream", {
    case SlideGeneralOrderedStream(VectorToOrderedStream(input, _), windowSize, stepSize, _) =>
      MapOrderedStream(VectorToOrderedStream.asFunction(), VectorToOrderedStream(SlideVector(input, windowSize, stepSize)))
  })

  def skipDropStream: Rule = Rule("moveDownSlideGeneralStream_skipDropStream", {
    case SlideGeneralOrderedStream(DropOrderedStream(input, firstElements, lastElements, _), windowSize, stepSize, _) =>
      val dropMargin = firstElements.ae.evalInt + lastElements.ae.evalInt
      val newWindowSize = windowSize.ae.evalInt + dropMargin
      val slidedInput = SlideGeneralOrderedStream(input, newWindowSize, stepSize)
      MapOrderedStream(DropOrderedStream.asFunction(Seq(None), Seq(firstElements, lastElements)), slidedInput)
  })

  def skipJoinStream: Rule = Rule("moveDownSlideGeneralStream_skipJoinStream", {
    case SlideGeneralOrderedStream(JoinOrderedStream(input, _), windowSize, stepSize, _) if {
      val innerLen = input.t.asInstanceOf[OrderedStreamTypeT].et.asInstanceOf[OrderedStreamTypeT].len
      windowSize.ae.evalInt % innerLen.ae.evalInt == 0 && stepSize.ae.evalInt % innerLen.ae.evalInt == 0
    } =>
      val innerLen = input.t.asInstanceOf[OrderedStreamTypeT].et.asInstanceOf[OrderedStreamTypeT].len
      val newWindowSize = windowSize.ae.evalInt / innerLen.ae.evalInt
      val newStepSize = stepSize.ae.evalInt / innerLen.ae.evalInt
      val slidedInput = SlideGeneralOrderedStream(input, newWindowSize, newStepSize)
      MapOrderedStream(JoinOrderedStream.asFunction(), slidedInput)
  })

  def skipMapFun: Rule = Rule("moveDownSlideGeneralStream_skipMapFun", {
    case SlideGeneralOrderedStream(MapOrderedStream(fun, input, _), windowSize, stepSize, _) =>
      MapOrderedStream(2, fun, SlideGeneralOrderedStream(input, windowSize, stepSize))
  })

  def skipMarker: Rule = Rule("moveDownSlideGeneralStream_skipMarker", {
    case SlideGeneralOrderedStream(Marker(input, text), windowSize, stepSize, _) =>
      Marker(SlideGeneralOrderedStream(input, windowSize, stepSize), text)
    case MapOrderedStream(body: LambdaT, Marker(input, text), _) if {
      var exprFound = false
      val innerMostBody = RulesHelper.getCurrentOrInnerMostLambda(body)
      innerMostBody.body match {
        case SlideGeneralOrderedStream(ParamUse(p3), windowSize, stepSize, _) if p3.id == innerMostBody.param.id =>
          exprFound = true
        case _ =>
      }
      exprFound
    } =>
      val innerMostBody = RulesHelper.getCurrentOrInnerMostLambda(body)
      val mapLevels = RulesHelper.getMapNDLevels(body)
      Marker(MapOrderedStream(mapLevels, innerMostBody, input), text)
  })

  def moveIntoLet: Rule = Rule("moveDownSlideGeneralStream_moveIntoLet", {
    case SlideGeneralOrderedStream(Let(arg, body, param, _), windowSize, stepSize, _) =>
      Let(arg, SlideGeneralOrderedStream(body, windowSize, stepSize), param)
    case MapOrderedStream(body: LambdaT, Let(arg, letBody, param, _), _) if {
      var exprFound = false
      val innerMostBody = RulesHelper.getCurrentOrInnerMostLambda(body)
      innerMostBody.body match {
        case SlideGeneralOrderedStream(ParamUse(p3), windowSize, stepSize, _) if p3.id == innerMostBody.param.id =>
          exprFound = true
        case _ =>
      }
      exprFound
    } =>
      val innerMostBody = RulesHelper.getCurrentOrInnerMostLambda(body)
      val mapLevels = RulesHelper.getMapNDLevels(body)
      Let(arg, MapOrderedStream(mapLevels, innerMostBody, letBody), param)
  })

  def skipMapVecJoinVec: Rule = Rule("moveDownSlideGeneralStream_skipMapVecJoinVec", {
    case SlideGeneralOrderedStream(VectorToOrderedStream(MapVector(ArchLambda(p1, JoinVector(ParamUse(p2), _), _), input, _), _), windowSize, stepSize, _) if p1.id == p2.id =>
      val slidedInput = SlideGeneralOrderedStream(VectorToOrderedStream(input), windowSize, stepSize)
      val slidedInputtc = TypeChecker.check(slidedInput)
      MapOrderedStream(
        {
          val param1 = ParamDef(slidedInputtc.t.asInstanceOf[OrderedStreamTypeT].et)
          ArchLambda(
            param1,
            MapOrderedStream(JoinVector.asFunction(), ParamUse(param1)
            )
          )
        }, slidedInputtc
      )
  })

  def mapFission: Rule = Rule("moveDownSlideGeneralStream_mapFission", {
    case MapOrderedStream(ArchLambda(p1, body, _), input, _) if {
      var exprFound = false
      var paramUses = 0
      body.visit {
        case SlideGeneralOrderedStream(ParamUse(p2), windowSize, stepSize, _) if p1.id == p2.id =>
          exprFound = true
        case ParamUse(p2) if p1.id == p2.id =>
          paramUses = paramUses + 1
        case _ =>
      }
      // prevent loops; check if this rewrite rule has already been applied
      val isSlideStreamOnly: Boolean = body match {
        case SlideGeneralOrderedStream(ParamUse(p2), windowSize, stepSize, _) if p1.id == p2.id => true
        case _ => false
      }
      !isSlideStreamOnly && exprFound && paramUses == 1
    } =>
      var windowSize: ArithTypeT = 0
      var stepSize: ArithTypeT = 0
      MapOrderedStream(
        {
          val param = ParamDef()
          ArchLambda(
            param,
            body.visitAndRebuild {
              case SlideGeneralOrderedStream(ParamUse(p2), ws, ss, _) if p1.id == p2.id =>
                windowSize = ws
                stepSize = ss
                ParamUse(param)
              case e => e
            }.asInstanceOf[Expr]
          )
        },
        MapOrderedStream(SlideGeneralOrderedStream.asFunction(Seq(None), Seq(windowSize, stepSize)), input)
      )
  })

  def mapFissionMapNDSlide: Rule = Rule("moveDownSlideGeneralStream_mapFissionMapNDSlide", {
    case MapOrderedStream(ArchLambda(p1, body, _), input, _) if {
      var exprFound = false
      var paramUses = 0
      body.visit {
        case MapOrderedStream(innerBody: LambdaT, ParamUse(p2), _) if
          p1.id == p2.id && {
            val innerMostBody = RulesHelper.getCurrentOrInnerMostLambda(innerBody)
            var innerExprFound = false
            innerMostBody.body match {
              case SlideGeneralOrderedStream(ParamUse(p3), windowSize, stepSize, _) if p3.id == innerMostBody.param.id =>
                innerExprFound = true
              case _ =>
            }
            innerExprFound
          } =>
          exprFound = true
        case ParamUse(p2) if p1.id == p2.id =>
          paramUses = paramUses + 1
        case _ =>
      }
      // prevent loops; check if this rewrite rule has already been applied
      val isMapNDSlideStmOnly: Boolean = body match {
        case MapOrderedStream(innerBody: LambdaT, ParamUse(p2), _) if
          p1.id == p2.id && {
            val innerMostBody = RulesHelper.getCurrentOrInnerMostLambda(innerBody)
            var innerExprFound = false
            innerMostBody.body match {
              case SlideGeneralOrderedStream(ParamUse(p3), windowSize, stepSize, _) if p3.id == innerMostBody.param.id =>
                innerExprFound = true
              case _ =>
            }
            innerExprFound
          } => true
        case _ => false
      }
      !isMapNDSlideStmOnly && exprFound && paramUses == 1
    } =>
      var windowSize: ArithTypeT = 0
      var stepSize: ArithTypeT = 0
      var innerLevels: Int = 0
      MapOrderedStream(
        {
          val param = ParamDef()
          ArchLambda(
            param,
            body.visitAndRebuild {
              case MapOrderedStream(innerBody: LambdaT, ParamUse(p2), _) if
                p1.id == p2.id && {
                  val innerMostBody = RulesHelper.getCurrentOrInnerMostLambda(innerBody)
                  var innerExprFound = false
                  innerLevels = RulesHelper.getMapNDLevels(innerBody)
                  innerMostBody.body match {
                    case SlideGeneralOrderedStream(ParamUse(p3), ws, ss, _) if p3.id == innerMostBody.param.id =>
                      windowSize = ws
                      stepSize = ss
                      innerExprFound = true
                    case _ =>
                  }
                  innerExprFound
                } =>
                ParamUse(param)
              case e => e
            }.asInstanceOf[Expr]
          )
        },
        MapOrderedStream(innerLevels + 1, SlideGeneralOrderedStream.asFunction(Seq(None), Seq(windowSize, stepSize)), input)
      )
  })

  def mapFusion: Rule = Rule("moveDownSlideGeneralStream_mapFusion", {
    case MapOrderedStream(ArchLambda(p1, body1, _), MapOrderedStream(ArchLambda(p2, body2, _),
    input, _), _)
      if {
        var exprFound = false
        body1.visit {
          case SlideGeneralOrderedStream(ParamUse(p3), windowSize, stepSize, _) if p1.id == p3.id =>
            exprFound = true
          case _ =>
        }
        exprFound
      } && {
        var exprFound = false
        body2.visit {
          case ParamUse(p4) if p2.id == p4.id =>
            exprFound = true
          case _ =>
        }
        exprFound
      } =>
      MapOrderedStream({
        val newParam1 = ParamDef()
        ArchLambda(
          newParam1,
          body1.visitAndRebuild {
            case ParamUse(usedParam1) if usedParam1.id == p1.id =>
              body2.visitAndRebuild {
                case ParamUse(usedParam2) if usedParam2.id == p2.id =>
                  ParamUse(newParam1)
                case e => e
              }.asInstanceOf[Expr]
            case e => e
          }.asInstanceOf[Expr]
        )
      }, input
      )
  })

  def mapFusionMapNDSlide: Rule = Rule("moveDownSlideGeneralStream_mapFusionMapNDSlide", {
    case MapOrderedStream(ArchLambda(p1, body1, _), MapOrderedStream(ArchLambda(p2, body2, _),
    input, _), _)
      if {
        var exprFound = false
        body1.visit {
          case MapOrderedStream(innerBody: LambdaT, ParamUse(p2), _) if
            p1.id == p2.id && {
              val innerMostBody = RulesHelper.getCurrentOrInnerMostLambda(innerBody)
              var innerExprFound = false
              innerMostBody.body match {
                case SlideGeneralOrderedStream(ParamUse(p3), windowSize, stepSize, _) if p3.id == innerMostBody.param.id =>
                  innerExprFound = true
                case _ =>
              }
              innerExprFound
            } =>
            exprFound = true
          case _ =>
        }
        exprFound
      } && {
        var exprFound = false
        body2.visit {
          case ParamUse(p4) if p2.id == p4.id =>
            exprFound = true
          case _ =>
        }
        exprFound
      } =>
      MapOrderedStream({
        val newParam1 = ParamDef()
        ArchLambda(
          newParam1,
          body1.visitAndRebuild {
            case ParamUse(usedParam1) if usedParam1.id == p1.id =>
              body2.visitAndRebuild {
                case ParamUse(usedParam2) if usedParam2.id == p2.id =>
                  ParamUse(newParam1)
                case e => e
              }.asInstanceOf[Expr]
            case e => e
          }.asInstanceOf[Expr]
        )
      }, input
      )
  })

  def mergeCounter2D: Rule = Rule("moveDownSlideGeneralStream_mergeCounter2D", {
    case SlideGeneralOrderedStream(CounterInteger(start, inc, loop, dim, rep, t), windowSize, stepSize, _) if
      dim.size == 2 && rep.size == 2 && rep.head.ae.evalInt == 0 && rep(1).ae.evalInt == 0 && loop.ae.evalInt == 0 =>
      val bitWidth = t.asInstanceOf[OrderedStreamTypeT].leafType.bitWidth
      val numOfStep = (dim(1).ae.evalInt - windowSize.ae.evalInt) / stepSize.ae.evalInt + 1
      val newStepSize = stepSize.ae * dim.head.ae
      val windowCounter = CounterInteger(0, inc, Seq(dim.head.ae, windowSize), Some(IntType(bitWidth)))
      val stepCounter = CounterInteger(start, newStepSize, Seq(numOfStep), Some(IntType(bitWidth)))
      MapOrderedStream(
        {
          val param1 = ParamDef(stepCounter.t.asInstanceOf[OrderedStreamTypeT].leafType)
          ArchLambda(
            param1,
            MapOrderedStream(
              2,
              {
                val param2 = ParamDef(windowCounter.t.asInstanceOf[OrderedStreamTypeT].leafType)
                ArchLambda(
                  param2,
                  TruncInteger(AddInt(Tuple(ParamUse(param1), ParamUse(param2))), bitWidth)
                )
              }, windowCounter
            )
          )
        }, stepCounter
      )

    case MapOrderedStream(ArchLambda(p1, SlideGeneralOrderedStream(ParamUse(p2), windowSize, stepSize, _), _),
    CounterInteger(start, inc, loop, dim, rep, t), _) =>
      val bitWidth = t.asInstanceOf[OrderedStreamTypeT].leafType.bitWidth
      val numOfStep = (dim.head.ae.evalInt - windowSize.ae.evalInt) / stepSize.ae.evalInt + 1
      val windowCounter = CounterInteger(0, inc, Seq(windowSize), Some(IntType(bitWidth)))
      val innerStepCounter = CounterInteger(start, stepSize, Seq(numOfStep), Some(IntType(bitWidth)))
      val outerStepCounter = CounterInteger(start, dim(0).ae, Seq(dim(1).ae), Some(IntType(bitWidth)))
      val stepCounter = MapOrderedStream(
        {
          val param1 = ParamDef(outerStepCounter.t.asInstanceOf[OrderedStreamTypeT].et)
          ArchLambda(
            param1,
            MapOrderedStream(
              {
                val param2 = ParamDef(innerStepCounter.t.asInstanceOf[OrderedStreamTypeT].et)
                ArchLambda(
                  param2,
                  TruncInteger(AddInt(Tuple(ParamUse(param1), ParamUse(param2))), bitWidth)
                )
              }, innerStepCounter
            )
          )
        }, outerStepCounter
      )

      val stepCountertc = TypeChecker.check(stepCounter)

      MapOrderedStream(
        2,
        {
          val param1 = ParamDef(stepCountertc.t.asInstanceOf[OrderedStreamTypeT].leafType)
          ArchLambda(
            param1,
            MapOrderedStream(
              {
                val param2 = ParamDef(windowCounter.t.asInstanceOf[OrderedStreamTypeT].leafType)
                ArchLambda(
                  param2,
                  TruncInteger(AddInt(Tuple(ParamUse(param1), ParamUse(param2))), bitWidth)
                )
              }, windowCounter
            )
          )
        }, stepCountertc
      )
  })

  /*
   *********************************************************************************************************************
    SlideGeneralOrderedStream to SlideVector Rules
   *********************************************************************************************************************
   */
  def skipTransposeVec: Rule = Rule("moveDownSlideGeneralStream_skipTransposeVec", {
    case SlideGeneralOrderedStream(VectorToOrderedStream(SplitVector(PermuteVector(JoinVector(input, _), permuteFun, _), chunkSize, _), _), windowSize, stepSize, _) if {
      var exprFound = false
      permuteFun.b.ae match {
        case Sum(IntDiv(v1, c1) :: Prod(Cst(c3) :: Mod(v2, c2) :: Nil) :: Nil) if v1 == v2 && c1 == c2 && c3.toInt == chunkSize.ae.evalInt =>
          exprFound = true
        case _ =>
      }
      exprFound
    } =>
      val slidedVecInput = MapVector(SlideVector.asFunction(Seq(None), Seq(windowSize, stepSize)), input)
      val slidedVecInputtc = TypeChecker.check(slidedVecInput)
      val joinIntermediate = MapVector(JoinVector.asFunction(), slidedVecInputtc)
      val transposedIntermediate = Transpose(joinIntermediate)
      val transposedIntermediatetc = TypeChecker.checkIfUnknown(transposedIntermediate, transposedIntermediate.t)
      MapOrderedStream(VectorToOrderedStream.asFunction(), VectorToOrderedStream(SplitVector(transposedIntermediatetc, windowSize)))
  })

  def moveUpMapConversion: Rule = Rule("moveUpMapConversion", {
    case MapOrderedStream(ArchLambda(p1, MapOrderedStream(convFun: LambdaT, body, _), _), SlideOrderedStream(input, windowSize, _), _) if
      RulesHelper.containsParam(body, p1) && {
        var exprFound = false
        convFun.body match {
          case Conversion(ParamUse(p1), t) if p1.id == convFun.param.id =>
            exprFound = true
          case Conversion(DropVector(Conversion(ParamUse(p1), _), _, _, _), _) if p1.id == convFun.param.id =>
            exprFound = true
          case _ =>
        }
        exprFound
      } =>
      MapOrderedStream(2, convFun, MapOrderedStream(ArchLambda(p1, body), SlideOrderedStream(input, windowSize)))
  })

  def skipStmToVec: Rule = Rule("MoveDownSlideStm_skipStmToVec", {
    case MapOrderedStream(ArchLambda(p1, OrderedStreamToVector(body, _), _), SlideOrderedStream(input, winSize, _), _) if RulesHelper.containsParam(body, p1) =>
      MapOrderedStream(OrderedStreamToVector.asFunction(), MapOrderedStream(ArchLambda(p1, body), SlideOrderedStream(input, winSize)))
  })
}
